
#ifndef _BGE_HLSL_
#define _BGE_HLSL_

const static bool enableHDR = true;

#include "shaders/common/hlsl.h"
#include "shaders/common/matrixUtils.h.hlsl"

#if defined(BGE_SHADERGEN)
uniform SamplerState defaultSampler2D : REGISTER(s0, space2);
uniform SamplerState defaultSamplerCube : REGISTER(s1, space2);
uniform SamplerState defaultSamplerRT : REGISTER(s2, space2);
uniform SamplerState defaultSamplerPoint : REGISTER(s3, space2);
#endif


static float M_HALFPI_F   = 1.57079632679489661923f;
static float M_PI_F       = 3.14159265358979323846f;
static float M_2PI_F      = 6.28318530717958647692f;


/// Calculate fog based on a start and end positions in worldSpace.
float computeSceneFog(  float3 startPos,
                        float3 endPos,
                        float fogDensity,
                        float fogDensityOffset,
                        float fogHeightFalloff )
{
   float f = length( startPos - endPos ) - fogDensityOffset;
   float h = 1.0 - ( endPos.z * fogHeightFalloff );
   return exp( -fogDensity * f * h );
}


/// Calculate fog based on a start and end position and a height.
/// Positions do not need to be in worldSpace but height does.
float computeSceneFog( float3 startPos,
                       float3 endPos,
                       float height,
                       float fogDensity,
                       float fogDensityOffset,
                       float fogHeightFalloff )
{
   float f = length( startPos - endPos ) - fogDensityOffset;
   float h = 1.0 - ( height * fogHeightFalloff );
   return exp( -fogDensity * f * h );
}


/// Calculate fog based on a distance, height is not used.
float computeSceneFog( float dist, float fogDensity, float fogDensityOffset )
{
   float f = dist - fogDensityOffset;
   return exp( -fogDensity * f );
}


/// Convert a float4 uv in viewport space to render target space.
float2 viewportCoordToRenderTarget( float4 inCoord, float4 rtParams )
{
   float2 outCoord = inCoord.xy / inCoord.w;
   outCoord = ( outCoord * rtParams.zw ) + rtParams.xy;
   return outCoord;
}


/// Convert a float2 uv in viewport space to render target space.
float2 viewportCoordToRenderTarget( float2 inCoord, float4 rtParams )
{
   float2 outCoord = ( inCoord * rtParams.zw ) + rtParams.xy;
   return outCoord;
}


/// The number of additional substeps we take when refining
/// the results of the offset parallax mapping function below.
///
/// You should turn down the number of steps if your needing
/// more performance out of your parallax surfaces.  Increasing
/// the number doesn't yeild much better results and is rarely
/// worth the additional cost.
///
#define PARALLAX_REFINE_STEPS 3

/// Performs fast parallax offset mapping using
/// multiple refinement steps.
///
/// @param texMap The texture map whos alpha channel we sample the parallax depth.
/// @param texCoord The incoming texture coordinate for sampling the parallax depth.
/// @param negViewTS The negative view vector in tangent space.
/// @param depthScale The parallax factor used to scale the depth result.
///
float2 parallaxOffset( sampler2D texMap, float2 texCoord, float3 negViewTS, float depthScale )
{
   float depth = tex2D( texMap, texCoord ).a;
   float2 offset = negViewTS.xy * ( depth * depthScale );

   for ( int i=0; i < PARALLAX_REFINE_STEPS; i++ )
   {
      depth = ( depth + tex2D( texMap, texCoord + offset ).a ) * 0.5;
      offset = negViewTS.xy * ( depth * depthScale );
   }

   return offset;
}

#if defined(BGE_SHADERGEN)
float2 parallaxOffset( Texture2D texMap, float2 texCoord, float3 negViewTS, float depthScale )
{
   float depth = texMap.Sample(defaultSampler2D, texCoord ).a;
   float2 offset = negViewTS.xy * ( depth * depthScale );

   for ( int i=0; i < PARALLAX_REFINE_STEPS; i++ )
   {
      depth = ( depth + texMap.Sample(defaultSampler2D, texCoord + offset ).a ) * 0.5;
      offset = negViewTS.xy * ( depth * depthScale );
   }

   return offset;
}
#endif


/// The maximum value for 16bit per component integer HDR encoding.
static const float HDR_RGB16_MAX = 100.0;

/// The maximum value for 10bit per component integer HDR encoding.
static const float HDR_RGB10_MAX = 4.0;

/// Encodes an HDR color for storage into a target.
float3 hdrEncode( float3 sample )
{
   #if defined( BGE_HDR_RGB16 )

      return sample / HDR_RGB16_MAX;

   #elif defined( BGE_HDR_RGB10 )

      return sample / HDR_RGB10_MAX;

   #else

      // No encoding.
      return sample;

   #endif
}

/// Encodes an HDR color for storage into a target.
float4 hdrEncode( float4 sample )
{
   return float4( hdrEncode( sample.rgb ), sample.a );
}

/// Decodes an HDR color from a target.
float3 hdrDecode( float3 sample )
{
   #if defined( BGE_HDR_RGB16 )

      return sample * HDR_RGB16_MAX;

   #elif defined( BGE_HDR_RGB10 )

      return sample * HDR_RGB10_MAX;

   #else

      // No encoding.
      return sample;

   #endif
}

/// Decodes an HDR color from a target.
float4 hdrDecode( float4 sample )
{
   return float4( hdrDecode( sample.rgb ), sample.a );
}

/// Returns the luminance for an HDR pixel.
float hdrLuminance( float3 sample )
{
   // There are quite a few different ways to
   // calculate luminance from an rgb value.
   //
   // If you want to use a different technique
   // then plug it in here.
   //

   ////////////////////////////////////////////////////////////////////////////
   //
   // Max component luminance.
   //
   //float lum = max( sample.r, max( sample.g, sample.b ) );

   ////////////////////////////////////////////////////////////////////////////
   // The perceptual relative luminance.
   //
   // See http://en.wikipedia.org/wiki/Luminance_(relative)
   //
   const float3 RELATIVE_LUMINANCE = float3( 0.2125, 0.7154, 0.0721 );
   float lum = dot( sample, RELATIVE_LUMINANCE );

   ////////////////////////////////////////////////////////////////////////////
   //
   // The average component luminance.
   //
   //const float3 AVERAGE_LUMINANCE = float3( 0.3333, 0.3333, 0.3333 );
   //float lum = dot( sample, AVERAGE_LUMINANCE );

   return lum;
}

float3 linearToGammaColor( float3 linearCol)
{
   //return pow(max(linearCol, 0), 1.0 / 2.233333333);
   return max(1.055 * pow(max(0, linearCol), 0.416666667) - 0.055, 0);
}

float toLinearColor( float srgb)
{
   //return pow(max(srgb, 0), 2.233333333);
   return srgb * (srgb * (srgb * 0.305306011 + 0.682171111) + 0.012522878);
}

float3 toLinearColor( float3 srgb)
{
   //return pow(max(srgb, 0), 2.233333333);
   return srgb * (srgb * (srgb * 0.305306011 + 0.682171111) + 0.012522878);
}

float4 toLinearColor( float4 srgba)
{
   return float4(toLinearColor(srgba.rgb), srgba.a);
}

float3 rgb2hsv(float3 c)
{
    float4 K = float4(0.0, -1.0 / 3.0, 2.0 / 3.0, -1.0);
    float4 p = lerp(float4(c.bg, K.wz), float4(c.gb, K.xy), step(c.b, c.g));
    float4 q = lerp(float4(p.xyw, c.r), float4(c.r, p.yzx), step(p.x, c.r));

    float d = q.x - min(q.w, q.y);
    float e = 1.0e-10;
    return float3(abs(q.z + (q.w - q.y) / (6.0 * d + e)), d / (q.x + e), q.x);
}

float3 hsv2rgb(float3 c)
{
    float4 K = float4(1.0, 2.0 / 3.0, 1.0 / 3.0, 3.0);
    float3 p = abs(frac(c.xxx + K.xyz) * 6.0 - K.www);
    return c.z * lerp(K.xxx, clamp(p - K.xxx, 0.0, 1.0), c.y);
}

/// Called from the visibility feature to do screen
/// door transparency for fading of objects.
void fizzle(float2 vpos, float visibility)
{
   // NOTE: The magic values below are what give us
   // the nice even pattern during the fizzle.
   //
   // These values can be changed to get different
   // patterns... some better than others.
   //
   // Horizontal Blinds - { vpos.x, 0.916, vpos.y, 0 }
   // Vertical Lines - { vpos.x, 12.9898, vpos.y, 78.233 }
   //
   // I'm sure there are many more patterns here to
   // discover for different effects.

   float2x2 m = { vpos.x, 0.916, vpos.y, 0.350 };
   clip( visibility - frac( determinant( m ) ) );
}

#if defined(BGE_USE_SCENE_CBUFFER)

#define MaterialFlag_diffuseMapUV1      (1 << 0)
#define MaterialFlag_overlayMapUV1      (1 << 1)
#define MaterialFlag_lightMapUV1        (1 << 2) // TODO
#define MaterialFlag_detailMapUV1       (1 << 3)
#define MaterialFlag_normalMapUV1       (1 << 4)
#define MaterialFlag_normalDetailMapUV1 (1 << 5)
#define MaterialFlag_opacityMapUV1      (1 << 6)
#define MaterialFlag_colorPaletteMapUV1 (1 << 7)
#define MaterialFlag_specularMapUV1     (1 << 8)
#define MaterialFlag_reflectivityMapUV1 (1 << 9)
#define MaterialFlag_emissiveMapUV1     (1 << 10)

cbuffer RenderPassConstBuffer : REGISTER(b0, space2)
{
    float4 ambient;
    float3 fogData;
    float accumTime;
    float4 fogColor;
    float4 oneOverFarplane;
    float4x4 worldToCamera;
    float4x4 cameraToScreen;
    float4x4 viewProj;
    float3 vEye;
    float _padding0;
    float3 eyePosWorld;
    float _padding1;
    float4x4 eyeMat;
    float3 diffuseEyePosWorld;
    float _padding2;
    float4 clipPlane0;
    float4 projectionParams;
    int4 viewportParams;
    float4 lightInfoBufferParams;
    float4 lightParams;
    float2 atlasScale;
    float2 _padding3;

    // sun PSSM
    float4x4 worldToLightProj;
    float4 scaleX;
    float4 scaleY;
    float4 offsetX;
    float4 offsetY;
    float4 atlasXOffset;
    float4 atlasYOffset;
    float4 farPlaneScalePSSM;
    float4 overDarkPSSM;
    float2 atlasScalePSSM;
    float2 fadeStartLength;
    float  shadowSoftness;
    float3 _padding4;
}

float4 alphaDouble(float4 val)
{
   val.w += val.w;
   return val;
}

float4 unpackColor(uint packedColor)
{
   return float4(
      ((packedColor >> 0) & 0x000000ff),
      ((packedColor >> 8) & 0x000000ff),
      ((packedColor >> 16) & 0x000000ff),
      ((packedColor >> 24))
   ) * (1.0 / 255.0);
}

// render pass textures
uniform TextureCube sceneCubeMap : REGISTER(t0, space2);
uniform Texture2D lightInfoBuffer : REGISTER(t1, space2);
uniform Texture2D ambientOcclusionBuffer : REGISTER(t2, space2);
uniform TextureCube sceneCubeSpecMap : REGISTER(t5, space2);
uniform Texture2D ShadowMap : REGISTER(t6, space2);
uniform Texture2D gTapRotationTex : REGISTER(t7, space2);

#define PrimFlag_BillBoardTransform       (1 << 0)
#define PrimFlag_BillBoardTransformZAxis  (1 << 1)
#define PrimFlag_OutputColor              (1 << 2)

   cbuffer cspPrimitive : register(b1)
   {
      float4x4 localTrans;
      float4 windParams;
      uint primOutputColor1;
      float primVisibility;
      uint primFlags;
      float cspPrimitive_padding;
   }

   cbuffer cspInstance : register(b2)
   {
      float4x4 uObjectTrans;
      float3 windDirAndSpeed;
      float uVisibility;
      float4 inLightId;

      uint4 instanceColors;
      float2 distanceFadeParams;
      float2 cspInstance_padding0;
   };

#if BGE_HLSL_MODEL < 6
   cbuffer cspPush : register(b3)
   {
      float  instanceBase;
      float3  cspPush_padding1;
   }
#else
   #define instanceBase 0
#endif

   struct InstanceData
   {
      float4x4 uObjectTrans;
      float3 windDirAndSpeed;
      float uVisibility;
      float4 lightIds;

      uint4 instanceColors;

      float2 distanceFadeParams;
      float2 InstanceData_padding0;
   };

   StructuredBuffer<InstanceData> instancesBuffer : register(t3); // space 4?
   #define getInstanceData(idx) instancesBuffer[idx]

   float vehicleReflectivityFallback(float val)
   {
      return pow(max(val, 0), 2.0);
   }


float2 environmentBRDFAprox( float dNV, float roughness )
{
    float coeffs[] = {-1, -0.0275, -0.572, 0.022, 1, 0.0425, 1.04, -0.04, -9.28, -1.04, 1.04};
    float4 c0 = float4( coeffs[0], coeffs[1], coeffs[2], coeffs[3] );
    float4 c1 = float4( coeffs[4], coeffs[5], coeffs[6], coeffs[7] );
    float4 r = roughness * c0 + c1;
    float a004 = min( r.x * r.x, exp2( coeffs[8] * dNV ) ) * r.x + r.y;
    float2 AB = float2( coeffs[9], coeffs[10] ) * a004 + r.zw;
    return float2(AB.x, AB.y);
}


float3 environmentBRDFAprox( float3 specularColor, float roughness, float dNV )
{
    float2 AB = environmentBRDFAprox(dNV, roughness);
    return specularColor * AB.x + AB.y;
}

#endif // SHADERGEN

#endif // _BGE_HLSL_
